"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const resolve_1 = require("resolve");
const lodash_1 = require("lodash");
const lang_1 = require("../utils/lang");
var Env;
(function (Env) {
    Env["Production"] = "production";
})(Env = exports.Env || (exports.Env = {}));
var Command;
(function (Command) {
    Command["Bundle"] = "webpack";
    Command["Serve"] = "webpack-dev-server";
})(Command = exports.Command || (exports.Command = {}));
/**
 * Main class of the library
 *
 * @export
 * @class Bento
 */
class Bento {
    /**
     * Creates an instance of Bento.
     * @param {Config} config
     * @memberof Bento
     */
    constructor(config) {
        /**
         * @private
         * @type {RuleDescriptor[]}
         * @memberof Bento
         */
        this.rules = [];
        /**
         * @private
         * @type {PluginCollection}
         * @memberof Bento
         */
        this.plugins = {};
        /**
         * @private
         * @type {Entry}
         * @memberof Bento
         */
        this.entry = {};
        /**
         * @private
         * @type {Features}
         * @memberof Bento
         */
        this.features = {
            sourceMap: true
        };
        this.config = config;
        this.resolve = path_1.resolve.bind(this, this.cwd);
        this.load();
    }
    /**
     * Factory method to create an instance of Bento
     *
     * @static
     * @param {Config} config
     * @param {string} [cwd]
     * @returns {T}
     * @memberof T
     */
    static create(config, cwd) {
        return new this(config, cwd);
    }
    /**
     * Define `entry` value for webpack config
     *
     * @param {string} name
     * @param {...string[]} files
     * @returns {this}
     * @memberof Bento
     */
    bundle(name, files) {
        if (lodash_1.isString(files))
            files = [files];
        this.entry[name] = files.map(file => file.replace("~/", `${this.config.homeDir}/`));
        return this;
    }
    /**
     * Allow all configuration to be modified before webpack configuration
     * is built
     *
     * @param {(manifest: {}) => void} [fn]
     * @returns {this}
     * @memberof Bento
     */
    tinker(fn) {
        const manifest = {
            rules: this.rules,
            plugins: this.plugins,
            entry: this.entry
        };
        fn(manifest);
        return this;
    }
    /**
     * Create a new RuleDescriptor by a given data and add it to rule list
     *
     * @param {string} ext  File extension
     * @param {Loader | Loader[]} loaders Loaders to use
     * @param {Condition | Condition[]} include Specify path inclusion
     * @returns {this}
     * @memberof Bento
     */
    addRule(ext, loaders, include) {
        if (lodash_1.isString(ext))
            ext = [ext];
        if (typeof loaders !== "function") {
            if (!Array.isArray(loaders))
                loaders = [loaders];
            loaders = loaders.map(loader => {
                if (typeof loader === "string")
                    return { loader };
                return loader;
            });
        }
        this.rules.push({
            ext: ext,
            include: include || [this.homeDir],
            loaders
        });
        return this;
    }
    /**
     * Convenient method to add multiple rules
     *
     * @param {RuleMap} map Map of file extensions and their creators
     * @returns {this}
     * @memberof Bento
     */
    addRules(map = {}) {
        for (let p in map) {
            this.addRule(p.replace(/\s+/g, "").split(","), map[p]);
        }
        return this;
    }
    /**
     * Add to plugin list a new PluginDescriptor by given data
     *
     * @param {string | Constructor<Plugin>} plugin Plugin constructor
     * @param {any[]} args Plugin constructor arguments
     * @param {string} env Env this plugin should be loaded to
     * @returns {this}
     * @memberof Bento
     */
    addPlugin(plugin, args = [], env) {
        let key = env || "all";
        if (!this.plugins[key])
            this.plugins[key] = [];
        this.plugins[key].push({
            plugin,
            args
        });
        return this;
    }
    /**
     * Convenient method to add multiple plugins
     *
     * @param {PluginMap} map
     * @param {string} env Env this plugin should be loaded to
     * @returns {this}
     * @memberof Bento
     */
    addPlugins(map = {}, env) {
        for (let p in map) {
            this.addPlugin(p, map[p], env);
        }
        return this;
    }
    /**
     * Export data to a standard consumable webpack config file
     *
     * @param {string | void} [env]
     * @returns {Promise<Configuration>}
     * @memberof Bento
     */
    export(env) {
        const select = lang_1.selector(env);
        const plugins = [].concat(this.plugins.all || [], env ? this.plugins[env] || [] : []);
        const config = {
            name: "bento",
            entry: this.entry,
            output: {
                path: this.outputDir,
                pathinfo: select({
                    default: true,
                    production: false
                }),
                filename: select({
                    default: "[name].js",
                    production: "[name].[chunkhash:8].js"
                }),
                publicPath: this.publicPath
            },
            devtool: select({
                default: this.features.sourceMap
                    ? "cheap-module-eval-source-map"
                    : null,
                production: this.features.sourceMap ? "source-map" : false
            }),
            resolve: {
                alias: {
                    "~": this.homeDir
                }
            },
            module: {
                rules: this.rules.map(desc => {
                    return {
                        test: new RegExp(`\\.(${desc.ext.join("|")})$`),
                        include: desc.include,
                        use: typeof desc.loaders === "function"
                            ? desc.loaders(env)
                            : desc.loaders.map(loader => {
                                return typeof loader === "function"
                                    ? loader(env)
                                    : loader;
                            })
                    };
                })
            },
            plugins: plugins
                .map(desc => {
                let Plugin;
                if (typeof desc.plugin === "function")
                    Plugin = desc.plugin;
                else {
                    const module = resolve_1.sync(desc.plugin, {
                        basedir: this.cwd
                    });
                    Plugin = require(module);
                }
                const args = typeof desc.args === "function"
                    ? desc.args(env)
                    : desc.args;
                return lodash_1.isArray(args) ? lang_1.instantiate(Plugin, args) : null;
            })
                .filter(Boolean)
        };
        return Promise.resolve(this.configure(config, env));
    }
    /**
     * Enable / disable available features
     *
     * @param {string} flag
     * @param {any} value
     * @returns {this}
     * @memberof Bento
     */
    set(flag, value) {
        this.features[flag] = value;
        return this;
    }
    /**
     * Find loader with a given `name`, `cb` will be triggered if found.
     *
     * @param {string} name
     * @param {({ loader, ext, include }) => void} cb
     * @memberof Bento
     */
    findLoader(name, cb) {
        this.rules.forEach(rule => {
            const trigger = loader => {
                if (loader.loader && loader.loader === name)
                    cb({ loader, ext: rule.ext, include: rule.include });
            };
            if (typeof rule.loaders === "function")
                return;
            rule.loaders.forEach(trigger);
        });
    }
    /**
     * Resolved homeDir
     *
     * @readonly
     * @protected
     * @type {string}
     * @memberof Bento
     */
    get homeDir() {
        return this.resolve(this.config.homeDir);
    }
    /**
     * Resolved outputDir
     *
     * @readonly
     * @protected
     * @type {string}
     * @memberof Bento
     */
    get outputDir() {
        return this.resolve(this.config.outputDir);
    }
    /**
     * Default publicPath
     *
     * @readonly
     * @protected
     * @type {string}
     * @memberof Bento
     */
    get publicPath() {
        return this.config.publicPath || "/";
    }
    /**
     * @readonly
     * @protected
     * @type {string}
     * @memberof Bento
     */
    get cwd() {
        return this.config.cwd || process.cwd();
    }
    /**
     * Placeholder method for subclass to add in rules & plugins
     *
     * @protected
     * @memberof Bento
     */
    load() { }
    /**
     * Placeholder method for subclass to overwrite `config`
     *
     * @protected
     * @param {Configuration} config
     * @param {string | void} env
     * @returns {Configuration}
     * @memberof Bento
     */
    configure(config, env) {
        return config;
    }
    /**
     * @readonly
     * @protected
     * @type {string}
     * @memberof StandardBento
     */
    get command() {
        return path_1.basename(process.argv[1]);
    }
}
exports.default = Bento;
//# sourceMappingURL=Bento.js.map