"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const webpack_1 = require("webpack");
const ExtractTextWebpackPlugin = require("extract-text-webpack-plugin");
const path_1 = require("path");
const lodash_1 = require("lodash");
const Bento_1 = require("../core/Bento");
const lang_1 = require("../utils/lang");
class StandardBento extends Bento_1.default {
    /**
     * @see Bento#load
     * @protected
     * @memberof StandardBento
     */
    load() {
        // set default addtional features flag
        this.set("extractCss", true).set("emitFiles", false);
        // SCRIPT
        this.addRule(["js", "jsx"], {
            loader: "babel-loader",
            options: {
                presets: [
                    [
                        "@babel/preset-env",
                        {
                            useBuiltIns: "usage"
                        }
                    ]
                ]
            }
        })
            .addPlugin("clean-webpack-plugin", [
            this.config.outputDir,
            {
                root: this.cwd,
                verbose: false
            }
        ])
            .addPlugin("webpack-manifest-plugin", [])
            .addPlugin("write-file-webpack-plugin", (env) => {
            let args = null;
            if (this.features.emitFiles === true)
                args = [];
            if (this.features.emitFiles instanceof RegExp)
                args = [
                    {
                        test: this.features.emitFiles
                    }
                ];
            return args;
        })
            .addPlugin(webpack_1.NamedChunksPlugin, [
                chunk => {
                if (chunk.name) {
                    return chunk.name;
                }
                return chunk.mapModules(m => {
                    return path_1.basename(m.request, path_1.extname(m.request));
                });
            }
        ])
            .addPlugin(webpack_1.NamedModulesPlugin, [])
            .addPlugin(webpack_1.DefinePlugin, (env) => {
            return [
                {
                    "process.env": {
                        NODE_ENV: JSON.stringify(env)
                    }
                }
            ];
        })
            .addPlugin(webpack_1.optimize.CommonsChunkPlugin, [
            {
                name: "vendor",
                minChunks(module) {
                    return /node_modules/.test(module.resource);
                }
            }
        ], Bento_1.Env.Production)
            .addPlugin(webpack_1.optimize.CommonsChunkPlugin, [
            {
                name: "runtime"
            }
        ], Bento_1.Env.Production)
            .addPlugin("uglifyjs-webpack-plugin", (env) => {
            return [
                {
                    sourceMap: this.features.sourceMap
                }
            ];
        }, Bento_1.Env.Production);
        // STYLE
        this.addRule("scss", (env) => {
            return ExtractTextWebpackPlugin.extract({
                publicPath: "./",
                fallback: {
                    loader: "style-loader",
                    options: {
                        sourceMap: this.features.sourceMap
                    }
                },
                use: [
                    {
                        loader: "css-loader",
                        options: {
                            sourceMap: this.features.sourceMap,
                            minimize: env === Bento_1.Env.Production
                                ? { preset: "default" }
                                : false
                        }
                    },
                    {
                        loader: "postcss-loader",
                        options: {
                            ident: "postcss",
                            plugins() {
                                const autoprefixer = require("autoprefixer");
                                return [autoprefixer()];
                            },
                            sourceMap: this.features.sourceMap
                        }
                    },
                    {
                        loader: "resolve-url-loader",
                        options: {
                            sourceMap: this.features.sourceMap
                        }
                    },
                    {
                        loader: "sass-loader",
                        options: {
                            sourceMap: this.features.sourceMap
                        }
                    }
                ]
            });
        })
            .addRule("css", [
            {
                loader: "style-loader",
                options: {
                    sourceMap: this.features.sourceMap
                }
            },
            {
                loader: "css-loader",
                options: {
                    sourceMap: this.features.sourceMap
                }
            }
        ], /node_modules/)
            .addPlugin(ExtractTextWebpackPlugin, (env) => {
            const select = lang_1.selector(env);
            return [
                {
                    filename: select({
                        default: "[name].css",
                        production: "[name].[contenthash:8].css"
                    }),
                    disable: !(this.features.extractCss && env === Bento_1.Env.Production)
                }
            ];
        });
        // FONTS & MEDIA
        this.addRule([
            "ttf",
            "eot",
            "woff",
            "woff2",
            "jpeg",
            "jpg",
            "png",
            "gif",
            "svg",
            "mp4",
            "webm"
        ], (env) => {
            const select = lang_1.selector(env);
            return {
                loader: "file-loader",
                options: {
                    name: select({
                        default: "[name].[ext]",
                        production: "[name].[hash:8].[ext]"
                    })
                }
            };
        }, [this.homeDir, /node_modules/]);
        // HTML
        if (this.html) {
            this.addPlugin("html-webpack-plugin", [this.html]);
        }
    }
    /**
     * @see Bento#configure
     * @protected
     * @memberof StandardBento
     */
    configure(config) {
        if (this.command !== Bento_1.Command.Serve)
            return config;
        config.devServer = {
            disableHostCheck: true,
            historyApiFallback: true,
            overlay: true
        };
        if (this.proxy)
            config.devServer.proxy = this.proxy;
        return config;
    }
    /**
     * @readonly
     * @protected
     * @type {({} | void)}
     * @memberof StandardBento
     */
    get html() {
        let output = null;
        if (lodash_1.isString(this.config.html)) {
            output = {
                filename: this.config.html
            };
        }
        else if (lodash_1.isObject(this.config.html)) {
            output = this.config.html;
        }
        return output;
    }
    /**
     * @readonly
     * @protected
     * @type {({} | void)}
     * @memberof StandardBento
     */
    get proxy() {
        let output = null;
        if (lodash_1.isString(this.config.proxy)) {
            output = {
                "**": this.config.proxy
            };
        }
        else if (lodash_1.isObject(this.config.proxy)) {
            output = this.config.proxy;
        }
        return output;
    }
}
exports.default = StandardBento;
//# sourceMappingURL=StandardBento.js.map